﻿using System;
using System.Drawing;
using System.Drawing.Imaging;
using System.Linq;
using System.Runtime.InteropServices;
using System.Windows.Forms;

namespace ImageTransporter
{
    public partial class ImageTransporter : Form
    {
        //マウスの押された位置
        private Point mouseDownPoint = Point.Empty;
        private Point mouseDownPointMoto = Point.Empty;
        private int hitTestIndex2;
      

        public ImageTransporter()
        {
            InitializeComponent();
            // ===> 追加(実験用コード) ===>
            var il = new ImageList() { ImageSize = new Size(32, 32) };
            var rect = new RectangleF(0F, 0F, 32F, 32F);
            using (var f = new Font("MS Gothic", 15F, FontStyle.Regular))
            {
                for (int i = 0; i < 10; i++)
                {
                    var bmp = new Bitmap(32, 32);
                    using (var g = Graphics.FromImage(bmp))
                    {
                        g.Clear(Color.LightPink);
                        g.DrawString(i.ToString(), f, Brushes.Black, rect);
                    }
                    il.Images.Add(bmp);
                }
            }
            listRoot.Items.Clear();
            listDest.Items.Clear();
            listRoot.LargeImageList = listDest.LargeImageList = il;
            listDest.AllowDrop = true;
            // <=== 追加(実験用コード) ===
        }

        /// <summary>
        /// フォームロード
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void ImageTransporter_Load(object sender, EventArgs e)
        {
            ListViewItem lvi = new ListViewItem();
            lvi.ImageIndex = 0;
            listRoot.Items.Add(lvi);

            lvi = new ListViewItem();
            lvi.ImageIndex = 1;
            listRoot.Items.Add(lvi);

            lvi = new ListViewItem();
            lvi.ImageIndex = 2;
            listRoot.Items.Add(lvi);

            TreeView tvw = new TreeView();
            tvw.ImageIndex = 0;


            SetControlSpacing(listRoot, 75, 100);


        }

        // 幅w、高さhのImageオブジェクトを作成
        Image createThumbnail(Image image, int w, int h)
        {
            Bitmap canvas = new Bitmap(w, h);

            Graphics g = Graphics.FromImage(canvas);
            g.FillRectangle(new SolidBrush(Color.White), 0, 0, w, h);

            float fw = (float)w / (float)image.Width;
            float fh = (float)h / (float)image.Height;

            float scale = Math.Min(fw, fh);
            fw = image.Width * scale;
            fh = image.Height * scale;

            g.DrawImage(image, (w - fw) / 2, (h - fh) / 2, fw, fh);
            g.Dispose();

            return canvas;
        }


        /// <summary>
        /// 半透明の画像を生成します。
        /// </summary>
        /// <param name="sourceImage"></param>
        /// <returns></returns>
        public static Image CreateAlphaImage(Image sourceImage)
        {
            int imageWidth = sourceImage.Width;
            int imageHeight = sourceImage.Height;

            // 新しいビットマップを用意
            Bitmap alphaImage = new Bitmap(imageWidth, imageHeight);

            using (Graphics g = Graphics.FromImage(alphaImage))
            {
                // ColorMatrixオブジェクトの作成
                ColorMatrix cm = new ColorMatrix();

                // ColorMatrixの行列の値を変更して、アルファ値が0.5に変更されるようにする
                cm.Matrix00 = 1;
                cm.Matrix11 = 1;
                cm.Matrix22 = 1;
                cm.Matrix33 = 1;//0.5F;
                cm.Matrix44 = 1;

                // ImageAttributesオブジェクトの作成
                ImageAttributes ia = new ImageAttributes();

                // ColorMatrixを設定
                ia.SetColorMatrix(cm);

                // ImageAttributesを使用して画像を描画
                g.DrawImage(sourceImage,
                            new Rectangle(0, 0, imageWidth, imageHeight),
                            0,
                            0,
                            imageWidth,
                            imageHeight,
                            GraphicsUnit.Pixel,
                            ia);
            }

            return alphaImage;
        }

        [DllImport("user32.dll", CharSet = CharSet.Auto, SetLastError = false)]
//        private static extern Int32 SendMessage(IntPtr hwnd, Int32 wMsg, IntPtr wParam, IntPtr lParam);
        private static extern Int32 SendMessage(IntPtr hwnd, Int32 wMsg, IntPtr wParam, IntPtr lParam);

        const int LVM_FIRST = 0x1000;
        const int LVM_SETICONSPACING = LVM_FIRST + 53;

        /// <summary>
        /// コントロール間隔を修正します。
        /// </summary>
        /// <param name="control"></param>
        /// <param name="x"></param>
        /// <param name="y"></param>
        public void SetControlSpacing(Control control, int x, int y)
        {
            SendMessage(control.Handle, LVM_SETICONSPACING, new IntPtr(0), new IntPtr(x * 65536 + y));
            control.Refresh();
    }



        // listRoot.MouseMove および listDest.MouseMove の双方に割り当てる
        private void listView_MouseMove(object sender, MouseEventArgs e)
        {
            if (mouseDownPoint == Point.Empty)
            {
                return;
            }
            Rectangle moveRect = new Rectangle(
            mouseDownPoint.X - SystemInformation.DragSize.Width / 2,
            mouseDownPoint.Y - SystemInformation.DragSize.Height / 2,
            SystemInformation.DragSize.Width,
            SystemInformation.DragSize.Height);
            if (moveRect.Contains(e.X, e.Y))
            {
                return;
            }

            ListView lbx = (ListView)sender;
            var items = lbx.SelectedItems.OfType<ListViewItem>().ToArray();
           // lbx.AllowDrop = false; // ★
            var dde = lbx.DoDragDrop(items, DragDropEffects.Move);
            //  lbx.AllowDrop = true; // ★
            if (dde == DragDropEffects.Move)
            {
                lbx.BeginUpdate();
                foreach (var item in items)
                {
                    lbx.Items.Remove(item);
                }
                lbx.EndUpdate();

                int ItemCount = lbx.Items.Count;
                int[] imageHairetu = new int[ItemCount];
                //並び替え後の状態を保存
                foreach (ListViewItem item in lbx.Items)
                {
                    imageHairetu[item.Index] = item.ImageIndex;
                    System.Diagnostics.Debug.WriteLine(item.ImageIndex + "/" + item.Index);
                }
                if (ItemCount != 0)
                {
                    //クリア
                    lbx.Items.Clear();
                    //再割り当ておよび表示
                    for (int i = 0; i < ItemCount; ++i)
                    {
                        ListViewItem lvi = new ListViewItem();
                        lvi.ImageIndex = imageHairetu[i];
                        lbx.Items.Add(lvi);
                    }
                }
                SetControlSpacing(lbx, 75, 100);
            }
            mouseDownPoint = Point.Empty;
        }

        // listRoot.DragDrop および listDest.DragDrop の双方に割り当てる
        private void listView_DragDrop(object sender, DragEventArgs e)
        {
            var items = (ListViewItem[])e.Data.GetData(typeof(ListViewItem[]));

            ListView lbx = (ListView)sender;
            Point pReleasesPoint = lbx.PointToClient(new Point(e.X, e.Y));

            var hitTestIndex = (lbx.HitTest(pReleasesPoint).Item ?? new ListViewItem()).Index;
            var nearestIndex = lbx.InsertionMark.NearestIndex(pReleasesPoint);
            //ドロップ位置のアイテムindexを取得　デフォルトは最終行
            int ReleasesItemIndex = lbx.Items.Count;

           

            //どちらか最下行の場合は、リストの最大数を採用する
            if (hitTestIndex == (lbx.Items.Count - 1) || nearestIndex == (lbx.Items.Count - 1))
            {
                ReleasesItemIndex = lbx.Items.Count;
            }
            //確実にヒットした場合は、そのindexを採用
            else if (hitTestIndex != -1)
            {
                //上に移動されていればそのまま、下なら+1
                if (mouseDownPointMoto.Y > e.Y) ReleasesItemIndex = hitTestIndex;
                else  ReleasesItemIndex = hitTestIndex+1;
            }
            //ヒットしなく、近似値が最下行の場合は、リストの最大数を採用する
            else if (hitTestIndex == -1 && nearestIndex == (lbx.Items.Count - 1))
            {
                ReleasesItemIndex = lbx.Items.Count;
            }
            //上記以外は、近似値値を採用する。
            else
            {
                //上に移動されていればそのまま、下なら+1
                if (mouseDownPointMoto.Y > e.Y) ReleasesItemIndex = nearestIndex;
                else ReleasesItemIndex = nearestIndex+1;
            }

            
            lbx.BeginUpdate();
            foreach (var item in items.Reverse())
            {
                lbx.Items.Insert(ReleasesItemIndex, (ListViewItem)item.Clone());

            }
            lbx.EndUpdate();

             SetControlSpacing(lbx, 75, 100);
        }

        private void listView_MouseUp(object sender, MouseEventArgs e)
        {
            mouseDownPoint = Point.Empty;
        }

        /// <summary>
        /// コピー先のドラッグで入ってきたイベント
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void listView_DragEnter(object sender, DragEventArgs e)
        {
            if (e.Data.GetDataPresent(typeof(ListViewItem[])))
            {
                e.Effect = DragDropEffects.Move;
            }
            else
            {
                e.Effect = DragDropEffects.None;
            }
        }

        /// <summary>
        /// コピー元リストのドラッグ開始イベント
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void listView_MouseDown(object sender, MouseEventArgs e)
        {
            //マウスの左ボタンだけが押されている時のみドラッグできるようにする
            if (e.Button == MouseButtons.Left)
            {

                //マウスの押された位置を記憶
                mouseDownPoint = new Point(e.X, e.Y);
                mouseDownPointMoto = Cursor.Position;
            }
            else
                mouseDownPoint = Point.Empty;
        }

   
    }
  
}
